/* Util.cpp : Utility function for DFHL

DFHL - Duplicate File Hard Linker, a small tool to gather some space
    from duplicate files on your hard disk
Copyright (C) 2004-2012 Jens Scheffler & Oliver Schneider & Hans Schmidts

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


#include "StdAfx.h"
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <Windows.h>
#include <stdarg.h>
#include "Util.h"
#include "FileSystem.h"

// Global Variables
// *******************************************

/** Flag to write error messages to standard output (2012-10-15  HaSchm) */
bool boNoStdErr = false;


/** Logging Level for the console output */
int logLevel = LOG_INFO;


void logError(LPCWSTR message, ...) {
    va_list argp;
    FILE *fiErrOut = stderr;
    if (boNoStdErr) fiErrOut = stdout;  /* 2012-10-15  HaSchm */

    fwprintf(fiErrOut, L"ERROR: ");
    va_start(argp, message);
    vfwprintf(fiErrOut, message, argp);
    va_end(argp);
    fwprintf(fiErrOut, L"\n");
}

void logError(DWORD errNumber, LPCWSTR message, ...) {
    va_list argp;
    FILE *fiErrOut = stderr;
    if (boNoStdErr) fiErrOut = stdout;  /* 2012-10-15  HaSchm */

    fwprintf(fiErrOut, L"ERROR: ");
    va_start(argp, message);
    vfwprintf(fiErrOut, message, argp);
    va_end(argp);
    LPWSTR msgBuffer = NULL;
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        errNumber,
        GetSystemDefaultLangID(),
        (LPWSTR)(&msgBuffer),
        0,
        NULL);
    fwprintf(fiErrOut, L" -> [%i] %s\n", errNumber, msgBuffer);
    LocalFree(msgBuffer);
}

void logErrorInfo(LPCWSTR message, ...) {  /* 2012-10-15  HaSchm */
    va_list argp;
    FILE *fiErrOut = stderr;
    if (boNoStdErr) fiErrOut = stdout;

    va_start(argp, message);
    vfwprintf(fiErrOut, message, argp);
    va_end(argp);
    fwprintf(fiErrOut, L"\n");
}

void logInfo(LPCWSTR message, ...) {
    if (logLevel <= LOG_INFO) {
        va_list argp;
        va_start(argp, message);
        vwprintf(message, argp);
        va_end(argp);
        wprintf(L"\n");
    }
}

void logVerbose(LPCWSTR message, ...) {
    if (logLevel <= LOG_VERBOSE) {
        va_list argp;
        wprintf(L"  ");
        va_start(argp, message);
        vwprintf(message, argp);
        va_end(argp);
        wprintf(L"\n");
    }
}
void logDebug(LPCWSTR message, ...) {
    if (logLevel <= LOG_DEBUG) {
        va_list argp;
        wprintf(L"    ");
        va_start(argp, message);
        vwprintf(message, argp);
        va_end(argp);
        wprintf(L"\n");
    }
}

void setLogLevel(int newLevel) {
    logLevel = newLevel;
}

LPWSTR cleanString(LPWSTR string) {
    string[0] = 0;
    return string;
}

void logFile(WIN32_FIND_DATA FileData) {
    wchar_t alcSize[UTLi_nStringifySize1024StrSize];

    INT64 i64Size = FileData.nFileSizeLow + ((INT64)MAXDWORD + 1) * FileData.nFileSizeHigh;
    logDebug(L"Found file \"%s\" (Size=%I64i=%s,%s%s%s%s%s%s%s%s%s%s%s%s)",
        FileData.cFileName,
        i64Size,
        UTL_lsStringifySize1024(alcSize, (UINT64)i64Size, L'.', /*boSkipLBlanks*/true, /*boRetB*/true),
        FILE_ATTRIBUTE_ARCHIVE      &FileData.dwFileAttributes?L"ARCHIVE ":L"",
        FILE_ATTRIBUTE_COMPRESSED   &FileData.dwFileAttributes?L"COMPRESSED ":L"",
        FILE_ATTRIBUTE_DIRECTORY    &FileData.dwFileAttributes?L"DIRECTORY ":L"",
        FILE_ATTRIBUTE_ENCRYPTED    &FileData.dwFileAttributes?L"ENCRYPTED ":L"",
        FILE_ATTRIBUTE_HIDDEN       &FileData.dwFileAttributes?L"HIDDEN ":L"",
        FILE_ATTRIBUTE_NORMAL       &FileData.dwFileAttributes?L"NORMAL ":L"",
        FILE_ATTRIBUTE_OFFLINE      &FileData.dwFileAttributes?L"OFFLINE ":L"",
        FILE_ATTRIBUTE_READONLY     &FileData.dwFileAttributes?L"READONLY ":L"",
        FILE_ATTRIBUTE_REPARSE_POINT&FileData.dwFileAttributes?L"REPARSE_POINT ":L"",
        FILE_ATTRIBUTE_SPARSE_FILE  &FileData.dwFileAttributes?L"SPARSE ":L"",
        FILE_ATTRIBUTE_SYSTEM       &FileData.dwFileAttributes?L"SYSTEM ":L"",
        FILE_ATTRIBUTE_TEMPORARY    &FileData.dwFileAttributes?L"TEMP ":L"");
}

static Statistics* instance;
static int instanceFlag;

Statistics* Statistics::getInstance() {
    if (instanceFlag != 1234) {
        instance = new Statistics();
        instanceFlag = 1234;
    }
    return instance;
}

Statistics::Statistics() {
    i64FileCompares = 0;
    i64HashCompares = 0;
    i64HashComparesHit1 = 0;
    i64HashComparesHit2 = 0;
    i64HashComparesHit3 = 0;
    /* i64FileMetaDataMismatch = 0; 2012-10-12  HaSchm */
    i64FileNameMismatch = 0;
    i64FileAttributeMismatch = 0;
    i64FileMTimeMismatch = 0;

    i64FileAlreadyLinked = 0;
    i64FileContentDifferFirstBlock = 0;
    i64FileContentDifferLater = 0;
    i64FileContentSame = 0;
    fileCompareProblems = 0;
    i64HashesCalculated = 0;
    directoryOpenProblems = 0; /* 2012-10-16  HaSchm */
    directoriesFound = 0;
    junctionsFound = 0;  /* 2012-10-16  HaSchm */
    i64FilesFound = 0;
    i64HSFilesFound = 0;
    i64BytesRead = 0;
    i64FilesOpened = 0;
    i64FilesClosed = 0;
    fileOpenProblems = 0;
    pathObjCreated = 0;
    pathObjDestroyed = 0;
    i64FileObjCreated = 0;
    i64FileObjDestroyed = 0;
    i64UnbufferedFileStreamObjCreated = 0;
    i64UnbufferedFileStreamObjDestroyed = 0;
    i64FileSystemObjCreated = 0;
    i64FileSystemObjDestroyed = 0;
    i64HardLinks = 0;
    i64HardLinksSuccess = 0;
    i64BytesSaved = 0;  /* 2012-10-16  HaSchm */
    i64CollectionObjCreated = 0;
    i64CollectionObjDestroyed = 0;
    i64ItemObjCreated = 0;
    i64ItemObjDestroyed = 0;
    i64ReferenceCounterObjCreated = 0;
    i64ReferenceCounterObjDestroyed = 0;
    i64SizeGroupObjCreated = 0;
    i64SizeGroupObjDestroyed = 0;
    i64SortedFileCollectionObjCreated = 0;
    i64SortedFileCollectionObjDestroyed = 0;
    i64DuplicateEntryObjCreated = 0;
    i64DuplicateEntryObjDestroyed = 0;
    i64DuplicateFileCollectionObjCreated = 0;
    i64DuplicateFileCollectionObjDestroyed = 0;
    i64DuplicateFileHardLinkerObjCreated = 0;
    i64DuplicateFileHardLinkerObjDestroyed = 0;
    i64FileTooSmall = 0;
    i64TotalSmallFileSize = 0;
}


ReferenceCounter::ReferenceCounter() {
    Statistics::getInstance()->i64ReferenceCounterObjCreated++;
    numberOfReferences = 1;
}

ReferenceCounter::~ReferenceCounter() {
    Statistics::getInstance()->i64ReferenceCounterObjDestroyed++;
}

void ReferenceCounter::addReference() {
    numberOfReferences++;
}

void ReferenceCounter::removeReference() {
    numberOfReferences--;
    // Delete the object if no further reference exists
    if (numberOfReferences <= 0) {
        switch (classType) {
        case CLASS_TYPE_FILE:
            delete (File*) this;
            break;
        case CLASS_TYPE_PATH:
            delete (Path*) this;
            break;
        default:
            delete this;
            break;
        }
    }
}

/**
 * Item of the collection. Each entry is of type Item
 */
class Item {
public:
    void* data;

    Item* next;

    Item(void* defaultData) {
        Statistics::getInstance()->i64ItemObjCreated++;
        data = defaultData;
        next = NULL;
    }

    ~Item() {
        Statistics::getInstance()->i64ItemObjDestroyed++;
    }
};

void Collection::append(void* data) {
    Item* newItem = new Item(data);
    if (root != NULL) {
        last->next = newItem;
        last = newItem;
    } else {
        root = last = newItem;
    }
    itemCount++;
    nextItem = root;
}

void Collection::push(void* data) {
    append(data);
}

void Collection::addAll(Collection* items) {
    if (items == NULL) {
        return;
    }
    void* item;
    while (item = items->pop()) {
        append(item);
    }
}

void* Collection::pop() {
    if (root != NULL) {
        Item* temp = root;
        nextItem = root = root->next;
        if (last == temp) {
            last = root;
        }
        itemCount--;
        void* result = temp->data;
        delete temp;
        return result;
    } else {
        return NULL;
    }
}

void* Collection::next() {
    if (nextItem != NULL) {
        Item* temp = nextItem;
        nextItem = nextItem->next;
        return temp->data;
    } else {
        return NULL;
    }
}

void* Collection::item(int index) {
    if (itemCount -1 >= index) {
        Item* temp = root;
        for (int i=0; i<index; i++) {
            temp = temp->next;
        }
        nextItem = temp->next;
        return temp->data;
    } else {
        return NULL;
    }
}

int Collection::getSize() {
    return itemCount;
}

Collection::Collection() {
    Statistics::getInstance()->i64CollectionObjCreated++;
    itemCount = 0;
    root = last= nextItem = NULL;
}

Collection::~Collection() {
    Statistics::getInstance()->i64CollectionObjDestroyed++;
    while (itemCount > 0)
        pop();
}

extern "C" {
  #define UTLi_nStringifySize1024TabelEntries 7u
  const UINT64 UTLi_au64Limits[UTLi_nStringifySize1024TabelEntries] = {
                        /*0*/    /* "   0 Bytes " or "   0 B" */
                        /*1*/    /* "   1 Byte  " or "   1 B" */
                    1023ui64,    /* "1023 Bytes " or "1023 B" */
                 1048570ui64,    /* "1023.99 KiB" */
              1073736581ui64,    /* "1023.99 MiB" */
           1099506259066ui64,    /* "1023.99 GiB" */
        1125894409284485ui64,    /* "1023.99 TiB" */
     1152915875107312762ui64,    /* "1023.99 PiB" */
    18446744073709551615ui64     /* "  16.00 EiB" */  /* 2**64-1 = 15.9999... EiB */
  };
  /* Offsets are integer result of floor(0.005*1024**i). Added (indirect)
     to u64Size so that result of integer division is effectively
     rounded up to two decimal places. */
  const UINT64 UTLi_au64ROffs[UTLi_nStringifySize1024TabelEntries] = {
                   0ui64,  /* Bytes (offs. not used) */
                   5ui64,  /* KiB, 0,005*1024    = 5,12 */
                5242ui64,  /* MiB, 0,005*1024**2 = 5242,88 */
             5368709ui64,  /* GiB, 0,005*1024**3 = 5368709,12 */
          5497558138ui64,  /* TiB, 0,005*1024**4 = 5497558138,88 */
       5629499534213ui64,  /* PiB, 0,005*1024**5 = 5629499534213,12 */
    5764607523034234ui64   /* EiB, 0,005*1024**6 = 5764607523034234,88 */
  }; /* ai64ROffs */

  /***************************************************************************

  UTL_hsStringifySize1024: Convert a u64 size to string with binary unit

  copyright (c) 2021 BSK Datentechnik GmbH, Hans Schmidts
  */
  char * UTL_hsStringifySize1024(char *hs, UINT64 u64Size,
    char cDecPoint, bool boSkipLBlanks, bool boRetB) {

    static const char aacUnits[UTLi_nStringifySize1024TabelEntries][4] = {
      "<B>"/*(not used)*/, "KiB", "MiB", "GiB", "TiB", "PiB", "EiB"};
    int i;
    unsigned uShift;

    if (hs != NULL) {
      /* special handling for values <1024 */
      if (u64Size <= 1023ui64) {
        if (u64Size == 1ui64) {
          strcpy(hs,"   1 Byte  ");
        }
        else {
          (void)sprintf(hs,"%4" PRIu64 " Bytes ",u64Size);
        }
        if (boRetB) {
          hs[6] = '\0';  /* 6=length("nnnn B"): Cut "Bytes" or "Byte" to "B\0" */
        }
      }
      else {
        /* find range and calculate divisor (shift) */
        uShift = 3u;  /* = div 8 */
        for (i=1;;++i) {
          if (u64Size<=UTLi_au64Limits[i]) {
            break;
          }
          uShift += 10u;  /* >> 10 = div 1024 */
        } /* for */
        u64Size = (
          ((u64Size >> 5u) + (UTLi_au64ROffs[i] >> 5u)) * 25u
          + ((((u64Size & 0x1Fui64) + (UTLi_au64ROffs[i] & 0x1Fui64)) * 25u + 16u) >> 5u)
          ) >> uShift;

        (void)sprintf(hs,"%6" PRIu64 "  %s",u64Size,aacUnits[i]);
        /* insert decimal point */
        hs[4+2] = hs[4+1];  /* 4=length("nnnn"): Move 2. decimal digit */
        hs[4+1] = hs[4+0];  /* 4=length("nnnn"): Move 1. decimal digit */
        hs[4] = cDecPoint;  /* 4=length("nnnn"): Insert decimal point */
      }
      if (boSkipLBlanks) {
        /* skip leading blanks */
        if (hs[1]==' ') {
          if (hs[2]==' ') {
            hs = &hs[3];  /* result is <blank><blank><blank>"n.nn "<unit> */
          }
          else {
            hs = &hs[2];  /* result is <blank><blank>"nn.nn "<unit> */
          }
        }
        else if (hs[0]==' ') {
          hs = &hs[1];  /* result is <blank>"nnn.nn "<unit> */
        }
        /* else: result is "nnnn.nn "<unit> */
      } /* boSkipLBlanks */
    } /* hs != NULL */
    return hs;
  } /* UTL_hsStringifySize1024 */


  /***************************************************************************

  UTL_lsStringifySize1024: Convert a u64 size to string with binary unit
  */
  wchar_t * UTL_lsStringifySize1024(wchar_t *ls, UINT64 u64Size,
    wchar_t lcDecPoint, bool boSkipLBlanks, bool boRetB) {

    char acRes[UTLi_nStringifySize1024StrSize];
    int i;
    if (ls != NULL) {
      /* convert to local char string */
      (void)UTL_hsStringifySize1024(acRes, u64Size, (char)lcDecPoint, /*boSkipLBlanks*/false, boRetB);
      /* copy to wide string */
      for (i=0;;++i) {
        if ((ls[i]=acRes[i])==L'\0') {
          break;
        }
      } /* for i */
      if (boSkipLBlanks) {
        /* skip leading blanks */
        if (ls[1]==L' ') {
          if (ls[2]==L' ') {
            ls = &ls[3];  /* result is <blank><blank><blank>"n.nn "<unit> */
          }
          else {
            ls = &ls[2];  /* result is <blank><blank>"nn.nn "<unit> */
          }
        }
        else if (ls[0]==L' ') {
          ls = &ls[1];  /* result is <blank>"nnn.nn "<unit> */
        }
        /* else: result is "nnnn.nn "<unit> */
      } /* boSkipLBlanks */
    } /* ls != NULL */
    return ls;
  } /* UTL_lsStringifySize1024 */
} /* C */
